"""Tests that dynamically generate output for use in the documentation.

NOTE While the files and functions are named differently, they are essentially
     still regular test functions.
     See pyproject.toml for the pytest-related settings that allow this.
"""

import pytest

from .._fixtures import *

# -----------------------------------------------------------------------------


def gen_figures(out_dir):
    """Generates figures to use in the documentation.

    The ``out_dir`` fixture provides the directory to which the files should be
    written. These are then available during the documentation build (if the
    corresponding environment variables are set accordingly).
    """
    # TODO Can dynamically generate output here, storing it in `out_dir`
    pass



"""Various tools"""

import re
from typing import List, Union

import numpy as np

# -----------------------------------------------------------------------------


def str2bool(val: str) -> bool:
    """Copy of strtobool from deprecated distutils package"""
    val = val.lower()
    if val in ("y", "yes", "t", "true", "on", "1"):
        return True
    elif val in ("n", "no", "f", "false", "off", "0"):
        return False
    raise ValueError(f"Invalid truth value {repr(val)}!")


def eval_simple_math_expr(s: str) -> Union[int, float]:
    """Evalautes simple mathematical expressions, given by strings.

    Supports: +, -, *, **, /, e-X, eX, inf, nan
    """
    # Remove spaces
    expr_str = s.replace(" ", "")

    # Parse some special strings
    if expr_str in ["nan", "NaN"]:
        return float("nan")

    # NOTE these will cause errors if emitted file is not read by python!
    elif expr_str in ["np.inf", "inf", "INF"]:
        return np.inf

    elif expr_str in ["-np.inf", "-inf", "-INF"]:
        return -np.inf

    # remove everything that might cause trouble -- only allow digits, dot, +,
    # -, *, /, and eE to allow for writing exponentials, and parentheses
    expr_str = re.sub(r"[^0-9eE\-.+\*\/\(\)]", "", expr_str)

    # Try to eval
    return eval(expr_str)


def listgen(
    *,
    from_range: list = None,
    unique: bool = False,
    sort: bool = True,
    append: list = None,
    remove: list = None,
) -> List[int]:
    """Generates a list of integer elements.

    Args:
        from_range (list, optional): range arguments to use as the basis of the
            list
        unique (bool, optional): Whether to ascertain uniqueness of elements
        sort (bool, optional): Whether to sort the list before returning
        append (list, optional): Additional elements to append to the list
        remove (list, optional): Elements to remove all occurrences of

    Returns:
        List[int]: The generated list
    """
    l = []
    if from_range:
        l += list(range(*from_range))

    if append:
        l += append

    if remove:
        for element_to_remove in list(set(remove)):
            while element_to_remove in l:
                l.remove(element_to_remove)

    if unique:
        l = list(set(l))

    if sort:
        l.sort()

    return l



"""This module implements custom YAML representer functions"""

from typing import Any, Callable, Union

from .yaml import (
    BaseRepresenter,
    Node,
    RepresenterFunc,
    add_representer,
    is_representer,
)

# -- Multi-purpose representers -----------------------------------------------


def represent_by_type(
    representer: BaseRepresenter,
    obj: Union[list, tuple, dict, Any],
    *,
    tag: str,
) -> Node:
    """A representer for simple types: sequence-like, mapping-like or scalar"""
    if isinstance(obj, (list, tuple)):
        return representer.represent_sequence(tag, list(obj))

    elif isinstance(obj, dict):
        return representer.represent_mapping(tag, obj)

    return representer.represent_scalar(tag, str(obj))


# -- Representer factories ----------------------------------------------------


def build_representer(
    simplify_type: Callable[[Any], Union[list, tuple, dict, Any]]
) -> RepresenterFunc:
    """Builds a representer that"""

    def representer_func(representer: BaseRepresenter, obj: Any, *, tag: str):
        return represent_by_type(
            representer,
            simplify_type(obj),
            tag=tag,
        )

    return representer_func


# -- Registration of representers ---------------------------------------------

add_representer(
    slice,
    build_representer(lambda o: [o.start, o.stop, o.step]),
)

add_representer(
    range,
    build_representer(lambda o: [o.start, o.stop, o.step]),
)



"""Defines and registers YAML constructors."""

import copy as _copy
import operator as _operator
import os
import platform
import sys
from functools import partial as _partial
from functools import reduce as _reduce
from typing import Any, Callable

import numpy as np
import ruamel.yaml

from . import tools
from .yaml import (
    BaseLoader,
    ConstructorFunc,
    Node,
    add_constructor,
    is_constructor,
)

# -- Multi-purpose constructors -----------------------------------------------
# ... can be specialized by applying `functools.partial`


def scalar_node_to_object(loader: BaseLoader, node: Node):
    """Attempts to convert the given scalar node to a null (Python None),
    a bool, an int, or a float object using the corresponding YAML constructor.
    If those conversions fail, constructs a scalar (which will typically result
    in a string being returned).
    """

    def construct_yaml_null(node) -> None:
        """Constructs a None from an appropriate YAML node.

        This custom constructor should not be necessary, but for some weird
        reason, the ruamel.yaml constructor from loader.construct_yaml_null
        *always* returns None, regardless of the value of the node. As we rely
        on errors being raised if construction fails, we need this custom
        constructor for the two explicitly allowed null values.
        """
        if node.value in ("~", "null"):
            return None
        raise ruamel.yaml.constructor.ConstructorError(
            f"expected null, but " f"got '{node.value}'"
        )

    for constructor in (
        loader.construct_yaml_bool,
        loader.construct_yaml_int,
        loader.construct_yaml_float,
        construct_yaml_null,
        loader.construct_yaml_str,
    ):
        try:
            return constructor(node)
        except Exception:
            pass

    # Fallback -- very difficult to reach
    return loader.construct_scalar(node)


def construct_from_func(
    loader: BaseLoader, node: Node, *, func: Callable, unpack: bool = True
) -> Any:
    """A constructor that constructs a scalar, mapping, or sequence from the
    given node and subsequently applies the given function on it.

    Args:
        loader: The selected YAML loader
        node: The node from which to construct a Python object
        func (Callable): The callable to invoke on the resulting
        unpack (bool, optional): Whether to unpack sequences or mappings into
            the ``func`` call
    """

    def invoke_func(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as exc:
            raise ruamel.yaml.constructor.ConstructorError(
                "Constructing Python object from a tagged YAML node failed: "
                f"Got a {type(exc).__name__}: {exc}\n\n"
                f"Constructor:           {func}\n"
                f"Positional arguments:  {args}\n"
                f"Keyword arguments:     {kwargs}\n\n"
                "Check that the syntax was correct and the arguments have the "
                "expected types."
            ) from exc

    if isinstance(node, ruamel.yaml.nodes.MappingNode):
        s = loader.construct_mapping(node, deep=True)
        if unpack:
            return invoke_func(**s)

    elif isinstance(node, ruamel.yaml.nodes.SequenceNode):
        s = loader.construct_sequence(node, deep=True)
        if unpack:
            return invoke_func(*s)

    else:
        s = scalar_node_to_object(loader, node)

    return invoke_func(s)


# -- Batch-registration of simple constructors --------------------------------

# Programmatically define and add constructors, which
# evaluate nodes directly during construction. Distinguish between those where
# sequence or mapping arguments are NOT to be unpacked and those where
# unpacking them as positional and/or keyword arguments makes sense.
_func_constructors_no_unpack = [
    # built-ins operating on iterables
    ("!any", any),
    ("!all", all),
    ("!min", min),
    ("!max", max),
    ("!sum", sum),
    ("!prod", lambda a: _reduce(_operator.mul, a, 1)),
    ("!sorted", lambda a: list(sorted(a))),
    ("!isorted", lambda a: list(sorted(a, reverse=True))),
    #
    # built-ins operating on scalars
    ("!abs", lambda v: abs(float(v))),
    ("!int", lambda v: int(float(v))),
    ("!round", lambda v: round(float(v))),
    #
    # working with paths
    ("!expanduser", os.path.expanduser),
    #
    # numpy
    ("!array", np.array),
    #
    # misc
    ("!str-to-bool", tools.str2bool),
    ("!deepcopy", _copy.deepcopy),
]

_func_constructors_unpack = [
    # simple python types
    ("!slice", slice),
    ("!range", range),
    #
    # from operators module
    ("!add", _operator.add),
    ("!sub", _operator.sub),
    ("!mul", _operator.mul),
    ("!truediv", _operator.truediv),
    ("!floordiv", _operator.floordiv),
    ("!mod", _operator.mod),
    ("!pow", lambda x, y, z=None: pow(x, y, z)),
    ("!not", _operator.not_),
    ("!and", _operator.and_),
    ("!or", _operator.or_),
    ("!xor", _operator.xor),
    ("!lt", _operator.lt),
    ("!le", _operator.le),
    ("!eq", _operator.eq),
    ("!ne", _operator.ne),
    ("!neq", _operator.ne),
    ("!ge", _operator.ge),
    ("!gt", _operator.gt),
    ("!negate", _operator.neg),
    ("!invert", _operator.invert),
    ("!contains", _operator.contains),
    ("!concat", lambda *a: _reduce(_operator.concat, a, [])),
    ("!format", lambda fstr, *a, **k: fstr.format(*a, **k)),
    ("!join", lambda jstr, elements: jstr.join(elements)),
    ("!split", lambda s, *a: s.split(*a)),
    #
    # conditionals
    ("!if-else", lambda cond, a, b: a if cond else b),
    (
        "!if-unix-else",
        lambda a, b: a if platform.system() in ("Linux", "Darwin") else b,
    ),
    (
        "!if-windows-else",
        lambda a, b: a if platform.system() == "Windows" else b,
    ),
    #
    # numpy
    ("!arange", lambda *a: [float(f) for f in np.arange(*a)]),
    ("!linspace", lambda *a: [float(f) for f in np.linspace(*a)]),
    ("!logspace", lambda *a: [float(f) for f in np.logspace(*a)]),
    #
    # working with paths
    ("!joinpath", os.path.join),
]


# Add all of the above as constructors by specializing the func-constructor
for tag, func in _func_constructors_unpack:
    add_constructor(tag, _partial(construct_from_func, func=func, unpack=True))

for tag, func in _func_constructors_no_unpack:
    add_constructor(
        tag, _partial(construct_from_func, func=func, unpack=False)
    )


# -- Specialized constructors -------------------------------------------------


@is_constructor("!getenv", aliases=("!env",))
def getenv(loader: BaseLoader, node: Node):
    """Retrieves an environment variable by name, optionally with fallback"""
    if isinstance(node, ruamel.yaml.nodes.SequenceNode):
        return os.environ.get(*loader.construct_sequence(node, deep=True))
    return os.environ[str(loader.construct_scalar(node))]


@is_constructor("!getboolenv", aliases=("!boolenv",))
def getboolenv(loader: BaseLoader, node: Node) -> bool:
    """Retrieves an environment variable by name, optionally with fallback,
    and evaluates it as a boolean."""
    envvar = getenv(loader, node)
    if isinstance(envvar, str):
        return tools.str2bool(envvar)
    return envvar


@is_constructor(
    "!expr",
    aliases=("!expression", "!compute"),
    hint="Check the expression syntax",
)
def expression(loader: BaseLoader, node: Node):
    """Constructor that evaluates strings of simple mathematical expressions"""
    expr_str = loader.construct_scalar(node)
    return tools.eval_simple_math_expr(expr_str)


@is_constructor("!listgen")
def listgen(loader: BaseLoader, node: Node):
    """Constructor for lists, where node can be a mapping or sequence"""
    if isinstance(node, ruamel.yaml.nodes.MappingNode):
        kwargs = loader.construct_mapping(node, deep=True)

    elif isinstance(node, ruamel.yaml.nodes.SequenceNode):
        kwargs = dict(from_range=loader.construct_sequence(node))

    else:
        raise TypeError(
            f"Expected mapping or sequence node for !listgen, but "
            f"got {type(node)}!"
        )

    return tools.listgen(**kwargs)



"""This module registers various YAML constructors and representers.

Furthermore, it defines a shared ``ruamel.yaml.YAML`` object that can be
imported and used for loading and storing YAML files using the representers and
constructors.
"""
from functools import partial as partial
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import ruamel.yaml

from .exceptions import add_yaml_error_hint

# -- Types --------------------------------------------------------------------

Node = ruamel.yaml.nodes.Node

BaseLoader = ruamel.yaml.loader.BaseLoader

BaseRepresenter = ruamel.yaml.representer.BaseRepresenter

# .............................................................................

RepresenterFunc = Callable[[BaseRepresenter, Any, str], Node]

ResolvedRepresenterFunc = Callable[[BaseRepresenter, Any], Node]

ConstructorFunc = Callable[[BaseLoader, Node], Any]


# -- YAML objects -------------------------------------------------------------
yaml_safe: ruamel.yaml.YAML = ruamel.yaml.YAML(typ="safe")
"""An explicitly 'safe' YAML object"""

yaml: ruamel.yaml.YAML = yaml_safe
"""The default YAML object of yayaml"""


# -- Registration of representers and constructors ----------------------------

_REPRESENTERS: Dict[type, ResolvedRepresenterFunc] = {}
"""All representers that have been added by yayaml"""

_CONSTRUCTORS: Dict[str, ConstructorFunc] = {}
"""All constructors that have been added by yayaml"""

# .............................................................................


def add_representer(
    t: type,
    representer: RepresenterFunc,
    *,
    tag: Optional[str] = None,
    _yaml: Optional[ruamel.yaml.YAML] = None,
):
    """Adds a representer function for the given type"""
    if tag is None:
        tag = f"!{t.__name__}"

    resolved_representer = partial(representer, tag=tag)
    _REPRESENTERS[t] = resolved_representer

    if _yaml is None:
        yaml_safe.representer.add_representer(t, resolved_representer)
    else:
        _yaml.representer.add_representer(t, resolved_representer)


def add_constructor(
    tag: str,
    constructor: ConstructorFunc,
    *,
    aliases: List[str] = None,
    hint: Union[str, Tuple[Callable[[Exception], bool], str]] = None,
    _yaml: Optional[ruamel.yaml.YAML] = None,
):
    """Adds a constructor function for the given tag and optional aliases."""
    _CONSTRUCTORS[tag] = constructor

    if _yaml is None:
        yaml_safe.constructor.add_constructor(tag, constructor)
    else:
        _yaml.constructor.add_constructor(tag, constructor)

    # May want to add an error hint for that tag
    if hint:
        if isinstance(hint, str):
            match_func = lambda e: tag in str(e)
        else:
            match_func, hint = hint
        add_yaml_error_hint(match_func, hint)

    # Also register aliases with the same constructor and hint
    if aliases:
        for alias in aliases:
            add_constructor(alias, constructor, hint=hint, _yaml=_yaml)


# .. Decorators ...............................................................


def is_representer(
    t: type,
    *,
    tag: Optional[str] = None,
    _yaml: Optional[ruamel.yaml.YAML] = None,
):
    """Decorator to mark a function as the representer for a certain type"""

    def wrapper(representer: RepresenterFunc):
        add_representer(t, representer, tag=tag, _yaml=_yaml)
        return representer

    return wrapper


def is_constructor(
    tag: str,
    *,
    aliases: List[str] = None,
    hint: Union[str, Tuple[Callable[[Exception], bool], str]] = None,
    _yaml: Optional[ruamel.yaml.YAML] = None,
):
    """Decorator to mark a function as being a constructor for the given tag"""

    def wrapper(constructor: ConstructorFunc):
        add_constructor(
            tag, constructor, aliases=aliases, hint=hint, _yaml=_yaml
        )
        return constructor

    return wrapper



"""Reading and writing YAML files"""

import io
import logging
import os
from typing import Any, Union

import ruamel.yaml

from .exceptions import (
    YAML_ERROR_HINTS,
    RepresenterError,
    raise_improved_exception,
)
from .yaml import yaml

log = logging.getLogger(__name__)

# -- Writing and reading from files -------------------------------------------


def load_yml(
    path: str,
    *,
    mode: str = "r",
    encoding: str = "utf8",
    improve_errors: bool = True,
    _yaml: ruamel.yaml.YAML = yaml,
) -> Any:
    """Deserializes a YAML file into a Python object.

    Uses the yayaml-internal ``ruamel.yaml.YAML`` object for loading and thus
    supports all registered constructors.

    Args:
        path (str): The path to the YAML file that should be loaded. A ``~`` in
            the path will be expanded to the current user's directory.
        mode (str, optional): Read mode for the file at ``path``
        encoding (str, optional): The encoding to use. Assumes UTF-8 as the
            default; however, if there is a ``UnicodeDecodeError``, will try
            again with ``encoding=None``, thus using the default encoding of
            the operating system.
        improve_errors (bool, optional): Whether to improve error messages that
            come from the call to ``yaml.load``. If true, the error message
            is inspected and hints are appended.
        _yaml (ruamel.yaml.YAML, optional): If given, will use this YAML object
            for loading. By default, the yayaml-internal one is used, which
            supports all registered constructors.

    Returns:
        Any: The result of the data loading. Typically, this will be a dict,
            but depending on the structure of the file, it may be some other
            type, including ``None``.
    """
    path = os.path.expanduser(path)
    log.debug("Loading YAML file... mode: %s, path:\n  %s", mode, path)

    with open(path, mode, encoding=encoding) as yaml_file:
        try:
            return _yaml.load(yaml_file)

        except (ruamel.yaml.reader.ReaderError, UnicodeDecodeError):
            return load_yml(
                path,
                mode=mode,
                encoding=None,
                improve_errors=improve_errors,
                _yaml=_yaml,
            )

        except Exception as exc:
            if not improve_errors:
                raise

            # Attempt raising a new and improved error message.
            # Will simply re-raise if that is not the case.
            raise_improved_exception(exc, hints=YAML_ERROR_HINTS)


def write_yml(
    d: Union[dict, Any],
    *,
    path: str,
    mode: str = "w",
    _yaml: ruamel.yaml.YAML = yaml,
):
    """Serialize an object using YAML and store it in a file.

    Uses the dantro-internal ``ruamel.yaml.YAML`` object for dumping and thus
    supports all registered representers.

    Args:
        d (dict): The object to serialize and write to file
        path (str): The path to write the YAML output to. A ``~`` in the path
            will be expanded to the current user's directory.
        mode (str, optional): Write mode of the file
    """
    path = os.path.expanduser(path)
    log.debug(
        "Dumping %s to YAML file... mode: %s, path:\n  %s",
        type(d).__name__,
        mode,
        path,
    )

    # Make sure the directory is present
    os.makedirs(os.path.dirname(path), exist_ok=True)

    with open(path, mode) as yaml_file:
        # Add the yaml '---' prefix, then dump
        yaml_file.write("---\n")
        _yaml.dump(d, stream=yaml_file)
        yaml_file.write("\n")


# -- Serialization ------------------------------------------------------------


def yaml_dumps(
    obj: Any,
    *,
    register_classes: tuple = (),
    _yaml: ruamel.yaml.YAML = yaml,
    **dump_params,
) -> str:
    """Serializes the given object using (by default) yayaml's YAML object.

    Args:
        obj (Any): The object to dump
        register_classes (tuple, optional): Additional classes to register
        _yaml (ruamel.yaml.YAML, optional): Which YAML object to use for
            dumping
        **dump_params: Dumping parameters, which will be used to set the
            attributes of the given YAML object.

    Returns:
        str: The output of serialization

    Raises:
        RepresenterError: On failure to serialize the given
            object
    """
    s = io.StringIO()
    y = _yaml

    # Register classes; then apply dumping parameters via object properties
    for Cls in register_classes:
        y.register_class(Cls)

    for k, v in dump_params.items():
        setattr(y, k, v)

    # Serialize
    try:
        y.dump(obj, stream=s)

    except Exception as err:
        raise RepresenterError(
            f"Could not serialize the given {type(obj)} object!"
        ) from err

    return s.getvalue()


def yaml_dumps_plain(*args, **kwargs):
    """Like :py:func:`.yaml_dumps` but will create a completely new
    ``ruamel.yaml.YAML`` object for dumping, thus not having any of the special
    constructors and representers registered.

    The aim of this function is to provide YAML dumping that is not dependent
    on any package configuration; all parameters can be passed here.

    In other words, this function does _not_ use the yayaml's YAML object
    for dumping but each time creates a new dumper with fixed settings.
    This reduces the chance of interference from elsewhere.
    Compared to the time needed for serialization in itself, the extra time
    needed to create the new ruamel.yaml.YAML object and register the classes
    is negligible.

    .. hint::

        To use yayaml's YAML object, use :py:func:`.yaml_dumps` instead.

    .. note::

        The new YAML object will not have _any_ additional representers
        available!
    """
    return yaml_dumps(*args, **kwargs, _yaml=ruamel.yaml.YAML())



"""Custom exceptions and exception handling"""

from typing import Callable, List, Tuple

import ruamel.yaml
from ruamel.yaml.constructor import ConstructorError
from ruamel.yaml.representer import RepresenterError

# -- Custom exception types ---------------------------------------------------
# Aliases for now:

RepresenterError = ruamel.yaml.representer.RepresenterError
ConstructorError = ruamel.yaml.constructor.ConstructorError


# -- Error hints --------------------------------------------------------------


def raise_improved_exception(
    exc: Exception,
    *,
    hints: List[Tuple[Callable[[Exception], bool], str]] = [],
) -> None:
    """Improves the given exception by appending one or multiple hint messages.

    The ``hints`` argument should be a list of 2-tuples, consisting of a unary
    matching function, expecting the exception as only argument, and a hint
    that is part of the new error message.
    """
    matching_hints = []
    for match_func, hint in hints:
        if match_func(exc):
            matching_hints.append(hint)

    if matching_hints:
        _hints = "\n".join(f"  - {h}" for h in matching_hints)
        raise type(exc)(
            str(exc) + f"\n\nHint(s) how to resolve this:\n{_hints}"
        ) from exc

    # Re-raise the active exception
    raise


# .............................................................................

YAML_ERROR_HINTS: List[Tuple[Callable[[Exception], bool], str]] = [
    (
        lambda e: "expected ',' or ']'" in str(e),
        "Did you include a space after the YAML tag defined in that line?",
    ),
    #
    # Default case, always true:
    (
        lambda e: True,
        "Read the error message above for details about the error location.",
    ),
]
"""These are evaluated by :py:func:`.raise_improved_exception`
and from within :py:func:`.load_yml`.

Entries are of the form ``(match function, hint string)`` and can also be
added via :py:func:`.add_yaml_error_hint`.
"""


def add_yaml_error_hint(match_func: Callable[[Exception], bool], hint: str):
    """Adds an error hint for YAML error messages."""
    # Take into account that the last match function will always be true in the
    # case of the YAML error hints
    YAML_ERROR_HINTS.insert(
        max(0, len(YAML_ERROR_HINTS) - 1),
        (match_func, hint),
    )



"""Sets up the yayaml package installation"""

from setuptools import find_packages, setup

# .. Dependency lists .........................................................

INSTALL_DEPS = [
    "numpy < 2.0",
    "ruamel.yaml",
]

# Dependencies for running tests and general development of utopya
TEST_DEPS = [
    "pytest",
    "pytest-cov",
    "pre-commit",
]

# Dependencies for building the utopya documentation
DOC_DEPS = [
    "sphinx>=5.3,<6",
    "sphinx-book-theme",
    "sphinx-togglebutton",
    "ipython>=7.0",
    "myst-parser[linkify]",
    "pytest",
]

# .............................................................................


def find_version(*file_paths) -> str:
    """Tries to extract a version from the given path sequence"""
    import codecs
    import os
    import re

    def read(*parts):
        """Reads a file from the given path sequence, relative to this file"""
        here = os.path.abspath(os.path.dirname(__file__))
        with codecs.open(os.path.join(here, *parts), "r") as fp:
            return fp.read()

    # Read the file and match the __version__ string
    file = read(*file_paths)
    match = re.search(r"^__version__\s?=\s?['\"]([^'\"]*)['\"]", file, re.M)
    if match:
        return match.group(1)
    raise RuntimeError("Unable to find version string in " + str(file_paths))


# .............................................................................

DESCRIPTION = "yayaml makes yaml nicer. yay!"

LONG_DESCRIPTION = """
The `yayaml` package provides extensions to `ruamel.yaml` that allow creating
some often-needed Python objects directly via YAML tags and making it easier
to represent custom objects when writing YAML files.
"""
# .............................................................................


setup(
    name="yayaml",
    #
    # Package information
    version=find_version("yayaml", "__init__.py"),
    #
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    #
    url="https://gitlab.com/blsqr/yayaml",
    author="Yunus Sevinchan",
    author_email="Yunus Sevinchan <yunussevinchan@gmail.com>",
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        #
        "Development Status :: 5 - Production/Stable",
        #
        "Topic :: Utilities",
    ],
    #
    # Package content
    packages=find_packages(exclude=("tests",)),
    data_files=[("", ["README.md", "LICENSE", "CHANGELOG.md"])],
    #
    # Dependencies
    install_requires=INSTALL_DEPS,
    extras_require=dict(
        test=TEST_DEPS,
        doc=DOC_DEPS,
        dev=TEST_DEPS + DOC_DEPS,
    ),
)



"""
yayaml: A package providing tools for working with YAML.

Specifically, it makes it simpler to add custom constructors or representers,
and adds a whole suite of constructors for basic python functions.

"""
# Public interface
from .constructors import construct_from_func, scalar_node_to_object
from .exceptions import (
    ConstructorError,
    RepresenterError,
    add_yaml_error_hint,
    raise_improved_exception,
)
from .io import load_yml, write_yml, yaml_dumps, yaml_dumps_plain
from .representers import build_representer
from .yaml import (
    add_constructor,
    add_representer,
    is_constructor,
    is_representer,
    yaml,
    yaml_safe,
)

__version__ = "0.2.0"
